/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.experlog.lite;

import java.io.DataInputStream;
import java.io.IOException;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.ServletException;
import javax.servlet.ServletInputStream;

/**
 * Extracts parameter values from "multipart; form-data"
 * contained in incoming HTTP servlet requests. For further
 * details concerning multipart requests see RFC 1867.
 */
public class MultipartRequest {
    
    /** End of line sequence in multipart data sections */
    final public static String fEOLN = "\r\n";
    /**
     * Sequence containing two end of line sequences in multipart data sections
     * thus creating an empty line.
     */
    final public static String fEmptyLine = fEOLN + fEOLN;
    /** Terminating sequence at end of multipart section */
    final public static String fDoubleDash = "--";
    
    private HttpServletRequest mRequest;
    private String             mBoundary;
    private byte[]             mDataBytes;
    private String             mDataString;

    /**
     * Creates a new String out of an array of bytes or out of null
     * @param bytes an array of bytes or null
     * @return      an empty string if bytes is null otherwise a new String
     *              out of bytes
     */
    final static public String newString(byte[] bytes) {
	if (bytes == null) {
	    return "";
	}
	return new String(bytes);
    }

    /**
     * A convenience constructor allowing size of multipart content of incoming request to be
     * of maximum size.
     * @param request HttpServletRequest containing multipart message
     * @exception     ServletException is thrown if thrown by this(request, -1);
     */
    public MultipartRequest(HttpServletRequest request) throws ServletException{
	this(request, -1);
    }

    /**
     * Creates new IncomingMultiPartHttpServlet out of <request>.
     * @param request HttpServletRequest containing multipart message
     * @param maxSize max. allowable size of multipart message. If smaller than 0 it is assumed to
     *                be Integer.MAX_VALUE
     * @exception     ServletException is thrown if either of the following occurs:
     *                <ul><li>incoming request does not have a multipart message
     *                <li>multipart message is too large too process
     *                <li>IOException occurs while reading multipart message
     *                </ul>
     * @see #MultipartRequest(HttpServletRequest, int)
     */
    public MultipartRequest(HttpServletRequest request, int maxSize) throws ServletException{
	mRequest = request;
	if (getContentType().indexOf("multipart") == -1) {
	    throw new ServletException("ServletException: content-type should contain \"multipart\"\n"
				       + "Instead content-type is: " + getContentType());
	}

	//open input stream from client to capture upload file
	DataInputStream in = null;
	try {
	    in = new DataInputStream(getDataInputStream());
	}
	catch (IOException ex){
	    throw new ServletException(ex.toString());
	}

	//get length of content data
	int formDataLength = getContentLength();

	//allocate a byte array to store content data
	mDataBytes = new byte[formDataLength];
	//read file into byte array
	int bytesRead = 0;
	int totalBytesRead = 0;
	while (totalBytesRead < formDataLength) {
	    try{
		bytesRead = in.read(mDataBytes, totalBytesRead, formDataLength - totalBytesRead);
	    }
	    catch (IOException ex){
		throw new ServletException(ex.toString());
	    }
	    totalBytesRead += bytesRead;
	}

	mDataString = newString(mDataBytes);

	String contentType = mRequest.getContentType();
	int lastIndex = contentType.lastIndexOf("=");
	mBoundary = contentType.substring(lastIndex + 1, contentType.length());
    }

    // ==================== Header data ==================== //

    /**
     * Extracts DataInputStream from #mRequest.
     * @return DataInputStream associtated with #mRequest
     */
    public ServletInputStream getDataInputStream() throws IOException {
	return mRequest.getInputStream();
    }

    /**
     * Extracts ContentLength from #mRequest.
     * @return ContentLength associtated with #mRequest.
     */
    public int getContentLength(){
	return mRequest.getContentLength();
    }

    /**
     * Extracts ContentType from #mRequest.
     * @return ContentType associtated with #mRequest.
     */
    public String getContentType(){
	return mRequest.getContentType();
    }

    /**
     * Getter method to access #mDataBytes.
     * @return mDataBytes Array of bytes containing byte content of original request.
     */
    public byte[] getByteContent(){
	return mDataBytes;
    }

    /**
     * Extracts boundary from ContentType.
     * @return ContentType associtated with #mRequest.
     */
    public String getBoundary(){
	return mBoundary;
    }

    /**
     * Gets any query string that is part of the HTTP request URI (behind the '?').
     * Same as the CGI variable QUERY_STRING.
     * @return query string that is part of this request's URI, or null if it contains no query string
     */
    public String getQueryString() {
	return mRequest.getQueryString();
    }

    /**
     * Get IP address of calling client host from embedded HttpServletRequest.
     * @return String with IP address of calling client host.
     */
    public String getRemoteAddr() {
	return mRequest.getRemoteAddr();
    }

    /**
     * To be consistent the name of this method should actually be
     * getParameter. However, getParameter will be used to
     * access the parameters included in the multipart body. The same
     * applies to the other methods for accessing parameter names and
     * values. Methods for accessing the HttpServletRequest parameters
     * include the word "Header".
     * @param name String containing name of requested parameter value
     * @return String containing header parameter value
     */
    public String getHeaderParameter(String name) {
	return mRequest.getParameter(name);
    }

    // ==================== Multipart data ==================== //

    /**
     * Returns a string containing value of the specified parameter, or null if
     * parameter does not exist.
     * @param name String containing the name of the requested parameter value.
     * @return String or null
     */
    public String getMultipartHeader(String name) {
	DataBlock dblock = getBinaryHeaderBlock(name);
	if (dblock == null) {
	    return null;
	}

	byte bytes[] = dblock.getBlock();
	return newString(bytes);
    }

    /**
     * Returns a string containing value of the specified parameter, or null if
     * parameter does not exist.
     * @param name String containing the name of the requested parameter value.
     * @return String or null
     */
    public String getMultipartValue(String name) {
	DataBlock dblock = getBinaryValueBlock(name);
	if (dblock == null) {
	    return null;
	}
	byte bytes[] = dblock.getBlock();
	return newString(bytes);
    }


    /**
     * Extracts the binary content of parameter if available.
     * @param name String containing name of parameter of which binary content is requested
     * @return String Containing substring of mDataString with requested binary content
     */
    public byte[] getFileValue(String name){
	DataBlock db = getBinaryFileValue(name);
	if (db == null) {
	    return null;
	}
	return db.getBlock();
    }

    /**
     * Extracts content-type from parameter block with the name specified if available.
     * @param name String containing name of parameter of which content-type is requested
     * @return String Containing requested content-type if name belongs to a file data block
     *                otherwise null
     */
    public String getFileContentType(String name){
	DataBlock db = getBinaryFileBlock(name);
	if (db == null) {
	    return null;
	}
	return db.getContentType();
    }

    /**
     * Extracts file name from parameter block with the name specified if available.
     * @param name String containing name of parameter of which file name is requested
     * @return String Containing requested file name if name belongs to a file data block
     *                otherwise null
     */
    public String getFileName(String name){
	DataBlock db = getBinaryFileBlock(name);
	if (db == null) {
	    return null;
	}
	return db.getFileName();
    }

    /**
     * Retrieves the file name contained in the DataBlock with the specified
     * name stripped of any references to its path.
     * @param name String containing name of data block of which file name is
     *             requested.
     * @return String containing file name extracted from DataBlock with given name
     */
    public String getShortFileName(String name){
	DataBlock db = getBinaryFileBlock(name);
	if (db == null) {
	    return null;
	}
	return getShortFileName(db);
    }

    /**
     * To be consistent the name of this method should actually be
     * getParameter. However, as the depatcher uses the getParameter
     * method to access the object path. The object path parameter, however,
     * is included in the multipart body of mRequest and is not really
     * a HttpServletRequest. Thus getParameter will be used to
     * access the parameters included in the multipart body. The same
     * applies to the other methods for accessing parameter names and
     * values. Methods for accessing the HttpServletRequest parameters
     * include the word "Header".
     * @param name String containing name of requested parameter value
     * @return String containing header parameter value
     * @see #getHeaderParameter
     */
    public String getParameter(String name) {
	return getMultipartValue(name);
    }

    // ==================== request area ==================== //

    /**
     * Retrieves URL used to contact server
     * @return String with URL used to send request
     */
    public String getRequestURL() {
	return new String(mRequest.getRequestURL());
    }

    /**
     * Retrieves scheme used to make request, such as ftp, http, https
     * @return String with scheme used to make request.
     */
    public String getScheme() {
	return mRequest.getScheme();
    }

    /**
     * Determines whether request was made using a safe channel or not.
     * @return true if channel was safe (eg https) otherwise false
     */
    public boolean isSecure() {
	return mRequest.isSecure();
    }

    /**
     * Retrieves name of server from incoming HttpServletRequest
     * @return String with name of server
     */
    public String getServerName() {
	return mRequest.getServerName();
    }

    /**
     * Retrieves port id of server from incoming HttpServletRequest
     * @return int with port id of server
     */
    public int getServerPort() {
	return mRequest.getServerPort();
    }


    /**
     * A multipart request consists of a general header and of parameters delineated
     * by the boundary defined in the request header. The boundary is also used to
     * separate header and parameters. Also each parameter can be split into a header
     * and a body separated by an empty line (ie two consecutive line feeds). Each
     * parameter is assigned a single form-data item. The name of a parameter block
     * is the name given to the form-data item and is contained in the content-disposition
     * header entry.
     * @param name       String containing the name of the requested form-data item
     *                   (ie parameter block).
     * @return DataBlock containing parameter block, position within the request
     *                   and its name. Returns null if DataBlock with specified
     *                   name cannot be found within request or if parameter
     *                   block is not well-formed (ie does not contain a empty
     *                   line between header and value).
     * @see #getBinaryHeaderBlock
     * @see #getBinaryValueBlock
     */
    private DataBlock getBinaryParameterBlock(String name) {
	// Parameter block starts here:
	String find = mBoundary + fEOLN + "Content-Disposition: form-data; name=\"" + name + "\"";
	int index = mDataString.indexOf(find);
	if (index == -1) {
	    return null;
	}

	// Parameter block ends here:
	int index1 = mDataString.indexOf(fEOLN + fDoubleDash + mBoundary, index + fEOLN.length());
	if (index1 == -1) {
	    return null;
	}

	// Parameter block should contain an empty line between index and index1
	int index2 = mDataString.indexOf(fEmptyLine, index) + fEmptyLine.length();
	if (index2 == -1 || index2 > index1) {
	    return null;
	}

	index2 = mDataString.indexOf(fEOLN, index) + fEOLN.length();
	int uploadSize = index1 - index2;
	if (uploadSize < 0) {
	    return null;
	}
	byte[] uploadBytes = new byte[uploadSize];
	for (int i = 0; i < uploadSize; i++) {
	    uploadBytes[i] = mDataBytes[index2 + i];
	}
	return new DataBlock(index2, index1, uploadBytes, name);
    }

    /**
     * Retrieves the header part from the specified parameter block contained
     * within a multipart request.
     * @param name       String containing name of parameter block of which
     *                   header data is requested.
     * @return DataBlock containing the header block of the specified parameter
     *                   block, its name as well as position within the total
     *                   request. Returns null if DataBlock with specified
     *                   name cannot be found within request or if parameter
     *                   block is not well-formed (ie does not contain a empty
     *                   line between header and value).
     * @see #getBinaryParameterBlock
     * @see #getBinaryValueBlock
     */
    private DataBlock getBinaryHeaderBlock(String name) {
	DataBlock dataBlock = getBinaryParameterBlock(name);
	if (dataBlock == null) {
	    return null;
	}

	String block = newString(dataBlock.getBlock());
	int index = block.indexOf(fEmptyLine) + fEmptyLine.length();
	if (index == -1) {
	    return null;
	}

	byte[] bytes = new byte[index];
	for (int i = 0; i < index; i++) {
	    bytes[i] = dataBlock.getBlock()[index + i];
	}

	int from = dataBlock.getFrom();
	return new DataBlock(from, from + index, bytes, name);
    }

    /**
     * Retrieves the value part from the specified parameter block contained
     * within a multipart request.
     * @param name       String containing name of parameter block of which
     *                   value data is requested.
     * @return DataBlock containing the value block of the specified parameter
     *                   block, its name as well as position within the total
     *                   request. Returns null if DataBlock with specified
     *                   name cannot be found within request or if parameter
     *                   block is not well-formed (ie does not contain a empty
     *                   line between header and value).
     * @see #getBinaryParameterBlock
     * @see #getBinaryHeaderBlock
     */
    private DataBlock getBinaryValueBlock(String name) {
	DataBlock dataBlock = getBinaryParameterBlock(name);
	if (dataBlock == null) {
	    return null;
	}

	String block = newString(dataBlock.getBlock());
	int index = block.indexOf(fEmptyLine) + fEmptyLine.length();
	if (index == -1) {
	    return null;
	}

	int size = block.length() - index;
	byte[] bytes = null;
	if (size > 0) {
	    bytes = new byte[size];
	    for (int i = 0; i < size; i++) {
		bytes[i] = dataBlock.getBlock()[index + i];
	    }
	}

	int from = dataBlock.getFrom();
	return new DataBlock(from + index, from + size, bytes, name);
    }

    /**
     * Multipart requests are used amonst other things for uploading and
     * downloading file content. This convennience method can therefore
     * be used for accessing file content if available.
     * @param name       String containing name of parameter block of which
     *                   file data is requested.
     * @return DataBlock containing the requested parameter block, its name
     *                   as well as position within the total request. Returns
     *                   null if DataBlock with specified name cannot be found
     *                   within request or if parameter block is not well-formed
     *                   (ie does not contain a empty line between header and value)
     *                   or if parameter block does not contain any file data.
     * @see #getBinaryParameterBlock
     * @see #getBinaryHeaderBlock
     * @see #getBinaryValueBlock
     * @see #getBinaryFileHeader
     * @see #getBinaryFileValue
     */
    private DataBlock getBinaryFileBlock(String name) {
	DataBlock dataBlock = getBinaryParameterBlock(name);
	if (dataBlock == null) {
	    return null;
	}

	String parameter = mDataString.substring(dataBlock.getFrom(), dataBlock.getTo());
	String filename = "filename=\"";
	int index = parameter.indexOf(filename);
	if (index == -1) {
	    return null;
	}

	int index1 = parameter.indexOf("\"" + fEOLN, index);
	if (index1 == -1) {
	    return null;
	}

	dataBlock.setIsFileBlock(true);
	dataBlock.setFileName(parameter.substring(index + filename.length(), index1));

	String contentType = fEOLN + "Content-Type: ";
	index = parameter.indexOf(contentType);
	if (index != -1) {
	    index1 = parameter.indexOf("\"" + fEOLN, index);
	    if (index1 != -1) {
		dataBlock.setContentType(parameter.substring(index + contentType.length(), index1));
	    }
	    else {
		dataBlock.setContentType("Content-Type: text/plain");
	    }
	}
	else {
	    dataBlock.setContentType("Content-Type: text/plain");
	}
	return dataBlock;
    }

    /**
     * Retrieves the header part of a file block.
     * @param name       String containing name of file block of which
     *                   header data is requested.
     * @return DataBlock containing the requested file header block, if available,
     *                   or null.
     * @see #getBinaryParameterBlock
     * @see #getBinaryHeaderBlock
     * @see #getBinaryValueBlock
     * @see #getBinaryFileHeader
     * @see #getBinaryFileValue
     */
    private DataBlock getBinaryFileHeader(String name) {
	DataBlock dataBlock = getBinaryFileBlock(name);
	return getBinaryFileHeader(dataBlock);
    }

    /**
     * @param dataBlock  DataBlock containing block of which header data is requested.
     * @return DataBlock containing the requested file header block, its name
     *                   as well as position within the total request. Returns
     *                   null if DataBlock with specified name cannot be found
     *                   within request or if parameter block is not well-formed
     *                   (ie does not contain a empty line between header and value)
     *                   or if parameter block does not contain any file data.
     * @see #getBinaryFileHeader(String)
     */
    private DataBlock getBinaryFileHeader(DataBlock dataBlock){
	if (dataBlock == null) {
	    return null;
	}

	String block = newString(dataBlock.getBlock());
	int index = block.indexOf(fEmptyLine) + fEmptyLine.length();
	if (index == -1) {
	    return null;
	}

	byte[] bytes = new byte[index];
	for (int i = 0; i < index; i++) {
	    bytes[i] = dataBlock.getBlock()[i];
	}

	int from = dataBlock.getFrom();
	DataBlock db = new DataBlock(from, from + index, bytes, dataBlock.getName());
	db.setIsFileBlock(dataBlock.isFileBlock());
	db.setFileName(dataBlock.getFileName());
	return db;
    }

    /**
     * Retrieves the value part of a file block.
     * @param dataBlock  DataBlock containing block of which value block is requested.
     * @return DataBlock containing the requested file value block, if available,
     *                   or null.
     * @see #getBinaryParameterBlock
     * @see #getBinaryHeaderBlock
     * @see #getBinaryValueBlock
     * @see #getBinaryFileHeader
     * @see #getBinaryFileValue
     */
    private DataBlock getBinaryFileValue(String name) {
	DataBlock dataBlock = getBinaryFileBlock(name);
	return getBinaryFileValue(dataBlock);
    }

    /**
     * @param name       String containing name of file block of which
     *                   value block is requested.
     * @return DataBlock containing the requested file value block, its name
     *                   as well as position within the total request. Returns
     *                   null if DataBlock with specified name cannot be found
     *                   within request or if parameter block is not well-formed
     *                   (ie does not contain a empty line between header and value)
     *                   or if parameter block does not contain any file data.
     * @see #getBinaryParameterBlock
     * @see #getBinaryHeaderBlock
     * @see #getBinaryValueBlock
     * @see #getBinaryFileHeader
     * @see #getBinaryFileValue
     */
    private DataBlock getBinaryFileValue(DataBlock dataBlock) {
	if (dataBlock == null) {
	    return null;
	}

	String block = newString(dataBlock.getBlock());
	int index = block.indexOf(fEmptyLine) + fEmptyLine.length();
	if (index == -1) {
	    return null;
	}

	byte[] bytes = null;
	int size = block.length() - index;
	if (size > 0) {
	    bytes = new byte[size];
	    for (int i = 0; i < size; i++) {
		bytes[i] = dataBlock.getBlock()[index + i];
	    }
	}

	int from = dataBlock.getFrom();
	DataBlock db = new DataBlock(from + index, from + size, bytes, dataBlock.getName());
	db.setIsFileBlock(dataBlock.isFileBlock());
	db.setFileName(dataBlock.getFileName());
	return db;
    }

    /**
     * Retrieves the file name contained in the given DataBlock object.
     * @param dataBlock dataBlock of which file name is requested.
     * @return String containing file name extracted from DataBlock
     *                with given name
     */
    private String getShortFileName(DataBlock dataBlock) {
	String name = dataBlock.getFileName();
	if (name == null) {
	    return "";
	}
	int index = name.lastIndexOf("/");
	if (index == -1) {
	    index = name.lastIndexOf("\\");
	    if (index == -1) {
		return name;
	    }
	}

	if (index == name.length() - 1) {
	    return "";
	}

	return name.substring(index + 1, name.length());
    }

} // class MultipartRequest


