/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.experlog.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.sql.Connection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Locale;
import java.util.Properties;
import java.util.StringTokenizer;

import org.apache.log4j.Category;
import org.experlog.db.DatasourceProperty;

/**
 * A class to give access to shop properties
 * one shopConfigurator by application
 */
public class AppConfiguration implements Configuration {

  Properties conf_;

  // JDBC configuration (deprecated)
  public String driverClass_ = "org.experlog.db.DummyDriver";
  public String user_ = "";
  public String password_ = "";
  public String jdbcUrl_ = "jdbc:openeas:default";
  public String resultSet_ = "";
  public int maxConn_ = 0;
  public int minConn_ = 0;

  // Cookie : if usecookies is false, cookietimeout is ignored
  public boolean usecookies_ = false; // default no cookie (URL rewriting)
  public int cookietimeout_ = 30*24*3600; // Default 30 days
  public String cookielist_[] = null; // cookies to store in HTTP header

  public String jdbcTestStmt_ = "select 1";
  public int txIsolation_ = AppConfiguration.INT_DEFAULT;
  public boolean gc_ = true;
  public int checkLevelObject_ = 0;

  public String dataSourceName_ = null;
  public String sqlEncodeSubst = null;
  public String sqlEncodeSeparator = null;

  // Shop configuration
  public String name_ = "openeas";
  public String adminPasswd_ = null;
  public String baseDir_ = "";
  public String rootDir_ = "";
  public String currency_ = "USD";
  public int timeout_ = 3600; // Default 1 hour
  public boolean servletsOnly_ = false;
  public String offtmpl_ = null;
  public String cacheFile_ = null;
  public String uploaddir_ = null;
  public long lifeTime = 600000; //ms
  public long sleepTime = 600000; //ms
  public int maxClients_ = 999999;
  //public String encodingValue_ = "ISO-8859-1";
  public String encodingValue_ = "UTF8";

  Locale locale_ = Locale.getDefault(); // Locale that may evolve (L_ cookie)
  Locale appLocale_ = Locale.getDefault(); // Locale as defined in config. file
  boolean useFormatLocale_ = false; // Use variable locale (L_) for number/date formats

  public Hashtable contentType_ = null;

  // console parameters
  public String mbeanserver = "openeas";

  public String delegatefile = "";
  public String conffilename = "";
  public Category logger;

  public Hashtable datasources = null;

	public AppConfiguration() {
	}
	
	public AppConfiguration(
		String name,
		Properties props,
		boolean onWindows,
		String conffile,
		String basedir) {

		logger = ESCategory.getCategory(name, "openeas.util");

		name_ = name;
		conf_ = props;
		conffilename = conffile;

		if (basedir != null)
			baseDir_ =
				basedir
					+ (basedir.endsWith(File.separator) ? "" : File.separator);

		String s;

		parseProperties(props);

		s = props.getProperty("app.delegateconfig");
		if (s != null) {
			delegatefile = s;
			try {
				Properties p = new Properties();
				FileInputStream f = new FileInputStream(delegatefile);
				p.load(f);
				parseProperties(p);
				f.close();
			} catch (FileNotFoundException e) {

			} catch (IOException f) {

			}
		}
	}

	public void parseProperties(Properties props) {

		String s;

		/**
		 * JDBC primary datasource configuration
		 */
		s = props.getProperty("jdbc.Driver");
		if (s != null)
			driverClass_ = s.trim();

		s = props.getProperty("jdbc.Url");
		if (s != null)
			jdbcUrl_ = expandString(s.trim());
		
		s = props.getProperty("jdbc.User");
		if (s != null)
			user_ = s.trim();
		s = props.getProperty("jdbc.Password");
		if (s != null)
			password_ = s.trim();

		s = props.getProperty("jdbc.ResultSet");
		if (s != null)
			resultSet_ = s.trim();

		String jdbcMaxConn = props.getProperty("jdbc.MaxConn");
		String jdbcMinConn = props.getProperty("jdbc.MinConn");
		
		minConn_ = computeMin(jdbcMinConn);
		maxConn_ = computeMax(jdbcMinConn, jdbcMaxConn);

		s = props.getProperty("jdbc.JdbcTestStmt");
		if (s != null)
			jdbcTestStmt_ = s.trim();

		s = props.getProperty("jdbc.GC");
		if (s != null)
			gc_ = s.trim().equalsIgnoreCase("true");

		s = props.getProperty("jdbc.CheckLevelObject");
		if (s != null) {
			try {
				checkLevelObject_ = Integer.parseInt(s.trim());
			} catch (NumberFormatException e) {
				logger.error(
					"WARNING: No jdbc.CheckLevelObject specified, default 0.");
				checkLevelObject_ = 0;
				;
			}
		}

		txIsolation_ = decodeTxIsolation(props.getProperty("jdbc.TxIsolation"));

		s = props.getProperty("jdbc.DataSource");
		if (s != null)
			dataSourceName_ = s.trim();

		/**
		 * end of jdbc primary datasource config
		 */

		s = props.getProperty("console.applicationoff");
		if (s != null)
			offtmpl_ = makePath(s.trim());

		s = props.getProperty("shop.CacheFile");
		if (s != null)
			cacheFile_ = makePath(s.trim());

		s = props.getProperty("shop.AdminPassword");
		if (s == null)
			adminPasswd_ = new String("admin");
		else
			adminPasswd_ = s.trim();

		s = props.getProperty("app.RootDir");
		if (s == null) // backward compat
		  s = props.getProperty("shop.RootDir");
		if (s != null)
			rootDir_ = makePath(s.trim());

		s = props.getProperty("shop.Currency");
		if (s != null)
			currency_ = s.trim();

		s = props.getProperty("app.Timeout");
		if (s == null) // backward compat
		  s = props.getProperty("shop.Timeout");
		if (s != null) {
			try {
				timeout_ = Integer.parseInt(s.trim());
			} catch (NumberFormatException e) {
				logger.error(
					"WARNING: app.Timeout syntax error, default 3600 sec.");
				timeout_ = 3600;
			}
		}

		s = props.getProperty("app.usecookies");
		if (s != null) {
			usecookies_ = s.trim().equalsIgnoreCase("true");
			s = props.getProperty("app.cookietimeout");
			if (s != null) {
				try {
					cookietimeout_ = Integer.parseInt(s.trim());
				} catch (NumberFormatException e) {
                                	logger.error(
                                        	"WARNING: app.cookietimeout syntax error, default is 30 days.");
                                	cookietimeout_ = 30*24*3600;
                        	}
			}
			s = props.getProperty("app.httpcookies");
			if (s != null) {
				StringTokenizer st = new StringTokenizer(s, ",");
				// The HTTP spec says 20 cookies max
				cookielist_ = new String[20];
				for(int i=0; i<20; i++) {
					if(st.hasMoreTokens())
						cookielist_[i] = st.nextToken().trim();
					else
						cookielist_[i] = null;
				}
			}
                        
		}

		// Locale: syntax = language_country[_variant]
		// Default is system locale
		s = props.getProperty("app.Locale");
		if (s == null) // backward compat
		  s = props.getProperty("shop.Locale");
		if (s != null) {
		  setLocale(s);
		  appLocale_ = locale_;
		}
		useFormatLocale_ = "true".equalsIgnoreCase(props.getProperty("app.UseLocaleInFormats"));

		s = props.getProperty("app.ServletsOnly");
		if (s != null)
			servletsOnly_ = Boolean.valueOf(s.trim()).booleanValue();

		s = props.getProperty("app.UploadDir");
		if (s != null)
			uploaddir_ = s.trim();

		s = props.getProperty("console.mbeanserver");
		if (s != null)
			mbeanserver = s.trim();

		s = props.getProperty("sqlencode.Substitutions");
		if (s != null)
			sqlEncodeSubst = s.trim();

		s = props.getProperty("jdbc.sqlencode.Substitutions");
		if (s != null)
			sqlEncodeSubst = s.trim();

		s = props.getProperty("sqlencode.Separator");
		if (s != null)
			sqlEncodeSeparator = s.trim();

		s = props.getProperty("jdbc.sqlencode.Separator");
		if (s != null)
			sqlEncodeSeparator = s.trim();

		s = props.getProperty("app.MaxClients");
		if (s == null) // backward compat
		  s = props.getProperty("shop.MaxClients");
		if (s != null)
			maxClients_ = Integer.parseInt(s.trim());
		
		s = props.getProperty("urlencode.charset");
		if (s != null)
			encodingValue_ = s.trim();
		
		s = props.getProperty("trace.level");
		if (s != null)
			ESCategory.setLevel(s.trim(), name_);
		else
			ESCategory.setLevel("ERROR", name_);

		s = props.getProperty("trace.file");
		if (s != null)
			ESCategory.setOutput(makePath(s.trim()), name_);

                        
		s = props.getProperty("jdbc.LifeTime");
		if (s != null) {
			lifeTime = Long.parseLong(s);	
		}

		s = props.getProperty("jdbc.SleepTime");
		if (s != null) {
			sleepTime = Long.parseLong(s);	
		}
                        
		/**
		 * Other jdbc datasources, if present
		 */

		s = props.getProperty("datasource.list");
		if (s != null) {
			datasources = new Hashtable();
			StringTokenizer tok = new StringTokenizer(s.trim(), ",");
			while (tok.hasMoreTokens()) {

				boolean dsok = true;
				String datasourceName = tok.nextToken().trim();
				s =
					props.getProperty(
						"datasource." + datasourceName + ".Driver");
				DatasourceProperty dsprop =
					new DatasourceProperty(name_ + "__" + datasourceName);
				if (s != null) {
					dsprop.setDriver(s.trim());
				} else {
					dsok = false;
					logger.error(
						"Driver for datasource "
							+ datasourceName
							+ " must be defined");
				}

				s = props.getProperty("datasource." + datasourceName + ".Url");
				if (s != null) {
					dsprop.setUrl(expandString(s.trim()));
				} else {
					dsok = false;
					logger.error(
						"Url for datasource "
							+ datasourceName
							+ " must be defined");
				}

				s = props.getProperty("datasource." + datasourceName + ".User");
				if (s != null) {
					dsprop.setUser(s.trim());
				} else {
					dsprop.setUser("");
				}

				s =
					props.getProperty(
						"datasource." + datasourceName + ".Password");
				if (s != null) {
					dsprop.setPassword(s.trim());
				} else {
					dsprop.setPassword("");
				}

				s =
					props.getProperty(
						"datasource." + datasourceName + ".ResultSet");
				if (s != null) {
					dsprop.setResultSet(s.trim());
				} else {
					dsprop.setResultSet("");
				}

				s =
					props.getProperty(
						"datasource." + datasourceName + ".TxIsolation");
				dsprop.setTxIsolation(decodeTxIsolation(s));

				String dsmaxConn =
					props.getProperty(
						"datasource." + datasourceName + ".MaxConn");
				String dsminConn =
					props.getProperty(
						"datasource." + datasourceName + ".MinConn");

				s =
					props.getProperty(
						"datasource."
							+ datasourceName
							+ ".sqlencode.Substitutions");
				if (s != null) {
					dsprop.setSqlEncodeSubst(s.trim());
				}

				s =
					props.getProperty(
						"datasource."
							+ datasourceName
							+ ".sqlencode.Separator");
                                                        
                                                        
		                s = props.getProperty("datasource."+ datasourceName+ ".LifeTime");
		                if (s != null) {
			                dsprop.setLifeTime(Long.parseLong(s));	
		                } else
			                dsprop.setLifeTime(lifeTime);	                                                        

		                s = props.getProperty("datasource."+ datasourceName+ ".SleepTime");
		                if (s != null) {
			                dsprop.setSleepTime(Long.parseLong(s));	
		                } else
			                dsprop.setSleepTime(sleepTime);	                                                        
                                                        
				if (s != null) {
					dsprop.setSqlEncodeSeparator(s.trim());
				}

				dsprop.setMinConn(computeMin(dsminConn));
				dsprop.setMaxConn(computeMax(dsminConn, dsmaxConn));

				// Only keep properly configured datasources
				if (dsok)
					datasources.put(name_ + "__" + datasourceName, dsprop);
			}
			
		}
		
		s = props.getProperty("contenttype.list");
				
		if (s != null) {
			contentType_ = new Hashtable();
			StringTokenizer atok = new StringTokenizer(s.trim(), ",");
			while (atok.hasMoreTokens()) {
				String typeName = atok.nextToken().trim();
				String typeValue = props.getProperty("contenttype."+typeName+".value");
				contentType_.put(typeName, typeValue);
			}
		}

	} // parseProperties() method



        private String expandString(String s) {
 	 
 	     int pos = s.indexOf("${WEBAPP_PATH}");
 	     if(pos >= 0) {
 	       StringBuffer b = new StringBuffer(pos > 0 ? s.substring(0, pos) : "");
 	       b.append(baseDir_.substring(0, baseDir_.length()-1));
 	       if(s.length() > pos+14) b.append(s.substring(pos+14));
 	       return b.toString();
 	     } else return s;
        }

	private String makePath(String s) {
		if (new File(s).isAbsolute())
			return s;
		else
			return baseDir_ + s;
	}

	private int computeMin(String minC) {
		int iminC = 0;

		if (minC == null) {
			return 1;
		} else {
			try {
				iminC = Integer.parseInt(minC.trim());
			} catch (NumberFormatException e) {
				logger.error("WARNING: No min of connection, not specified.");
			}
			return iminC;
		}
	}

	private int computeMax(String minC, String maxC) {
		int iminC = 0;
		int imaxC = 0;

		if (maxC != null) {
			try {
				imaxC = Integer.parseInt(maxC.trim());
			} catch (NumberFormatException e) {
				logger.error("WARNING: No min of connection, not specified.");
			}
			return imaxC;
		} else {
			if (minC == null) {
				return 1;
			} else {
				try {
					iminC = Integer.parseInt(minC.trim());
				} catch (NumberFormatException e) {
					logger.error(
						"WARNING: No min of connection, not specified.");
				}
				return iminC;
			}
		}
		//	return 1;
	}

	public String getSqlEncodeSubst(String datasourcename) {
		if (datasourcename == null)
			return sqlEncodeSubst;
		else {
			if (datasources != null) {
				Iterator iter = datasources.values().iterator();
				while (iter.hasNext()) {
					DatasourceProperty dsprop =
						(DatasourceProperty) iter.next();
					if (dsprop
						.getName()
						.compareTo(name_ + "__" + datasourcename)
						== 0)
						return dsprop.getSqlEncodeSubst();
				}
			}
		}
		return null;
	}

	public String getSqlEncodeSeparator(String datasourcename) {
		if (datasourcename == null)
			return sqlEncodeSeparator;
		else {
			if (datasources != null) {
				Iterator iter = datasources.values().iterator();
				while (iter.hasNext()) {
					DatasourceProperty dsprop =
						(DatasourceProperty) iter.next();
					if (dsprop
						.getName()
						.compareTo(name_ + "__" + datasourcename)
						== 0)
						return dsprop.getSqlEncodeSeparator();
				}
			}
		}
		return null;
	}

	public Hashtable getDatasources() {
		return datasources;
	}

	// JDBC configuration
	public String getJdbcDriverClass() {
		return driverClass_;
	}

	public String getDataSourceName() {
		return dataSourceName_;
	}

	public String getJdbcUser() {
		return user_;
	}

	public String getJdbcPassword() {
		return password_;
	}

	public String getJdbcUrl() {
		return jdbcUrl_;
	}

	public String getResultSet() {
		return resultSet_;
	}

	public int getMaxConn() {
		return maxConn_;
	}

	public int getMinConn() {
		return minConn_;
	}

	public String getJdbcTestStmt() {
		return jdbcTestStmt_;
	}

	public boolean isGC() {
		return gc_;
	}

	public int getCheckLevelObject() {
		return checkLevelObject_;
	}

	public String getDriverClass() {
		return driverClass_;
	}

	public String getMbeanserver() {
		return mbeanserver;
	}

	public String getDelegatefile() {
		return delegatefile;
	}

	public String getConffilename() {
		return conffilename;
	}

	public int getMaxClients() {
	    return maxClients_;
	}
	
	public String getEncodingValue() {
	    return encodingValue_;
	}
	
        public long getLifeTime() {
		return lifeTime;	
	}
	
        public long getSleepTime() {
		return sleepTime;	
	}
        
	public void setJdbcDriverClass(String driverclass) {
		driverClass_ = driverclass;
	}

	public void setDataSourceName(String datasourcename) {
		dataSourceName_ = datasourcename;
	}

	public void setJdbcUser(String user) {
		user_ = user;
	}

	public void setJdbcPassword(String password) {
		password_ = password;
	}

	public void setJdbcUrl(String jdbcurl) {
		jdbcUrl_ = jdbcurl;
	}

	public void setMaxConn(int maxconn) {
		maxConn_ = maxconn;
	}

	public void setMinConn(int minconn) {
		minConn_ = minconn;
	}

	public void setJdbcTestStmt(String jdbcteststmt) {
		jdbcTestStmt_ = jdbcteststmt;
	}

	public void setGC(boolean gc) {
		gc_ = gc;
	}

	public void setCheckLevelObject(int checklevelobject) {
		checkLevelObject_ = checklevelobject;
	}

	public void setDriverClass(String driverclass) {
		driverClass_ = driverclass;
	}

	public void setTimeout(int time) {
		timeout_ = time;
	}

	public void setDelegatefile(String dfile) {
		delegatefile = dfile;
		conf_.setProperty("app.delegateconfig", dfile);
	}

	public void setMaxClients(int mc) {
	    maxClients_ = mc;
	}
	
	public void setEncodingValue(String encodingValue) {
	    encodingValue_ = encodingValue;
	}
	
	public String getContentType(String atype) {
		String ct = (String)contentType_.get(atype);
		return (ct == null ? "text/html" : ct);
	}
	
	
	// Shop configuration
	/**
	 * Get this shop's name
	 */
	public String getName() {
		return name_;
	}

	/**
	 * Get this shop's administrator password
	 */
	public String getAdminPassword(String id) {

		if (id == null)
			return adminPasswd_;

		// Do not return a default value here... would be a security hole!
		// An unregistered admin ID would cause the default value to be returned
		// as a password... so anyone could log in with the default password
		try {
			return get("shop.AdminPassword." + id);
		} catch (Exception e) {
			return null;
		}
	}

	/**
	 * Get this shop's root directory (shop.RootDir property)
	 */
	public String getRootDir() {
		return rootDir_;
	}

	/**
	 * Get this shop's currency unit (shop.Currency property)
	 */
	public String getCurrency() {
		return currency_;
	}

	/**
	 * Get the Locale
	 */
	public Locale getLocale() {
		return locale_;
	}
	
	public Locale getFormatLocale() {
	  return (useFormatLocale_ ? locale_ : appLocale_);
	}
	
	public void setLocale (String s) {
	  if(s == null) return;
	  s = s.trim();
	  if(s.length() < 1) {
	    try {
	      s = get("app.Locale");
	      if (s == null) s = get("shop.Locale"); // backward compat
	    } catch(Exception e) { s = null; }
	    if(s == null) {
	      locale_ = Locale.getDefault();
	      return;
	    }
	  }
	  
	  StringTokenizer st = new StringTokenizer(s, "_");
	  String llang = "en", lcntr = null;
	  if (st.hasMoreTokens()) { // Should always happen if not null!
	    llang = st.nextToken();
	  }
	  if (st.hasMoreTokens()) {
	    lcntr = st.nextToken();
	  }
	  if (st.hasMoreTokens()) {
	    locale_ = new Locale(llang, lcntr, st.nextToken());
	  } else if(lcntr != null) {
	    locale_ = new Locale(llang, lcntr);
	  } else {
	    locale_ = new Locale(llang);
	  }
	}

	public int getTimeout() {
		return timeout_;
	}

	public boolean servletsOnly() {
		return servletsOnly_;
	}

    /**
     * Enable exchanges between the HttpSession and the openeas "cookie"
     * To interact with external servlets or JSPs in the same webapp
     */
    public boolean useCookies() {
      return usecookies_;
    }

    /**
     * Cookie timeout is significant only if usecookies is set
     */
    public int getCookieTimeout() {
      return cookietimeout_;
    }

    /**
     * Check if name corresponds to a variable that should be stored
     * in the HTTP header cookie.
     */
    public boolean storeInCookie(String name) {
      if(cookielist_ == null) return false;
      // The HTTP spec says 20 cookies max
      for(int i=0; i<20 && cookielist_[i] != null; i++) {
        if(cookielist_[i].equals(name)) return true;
      }
      return false;
    }

	public String get(String pname) throws java.sql.SQLException {
		String s = conf_.getProperty(pname);
		return (s == null ? null : s.trim());
	}

	public String get(String pname, String value)
		throws java.sql.SQLException {
		String s = conf_.getProperty(pname, value);
		return (s == null ? null : s.trim());
	}

	public double getDouble(String pname) throws IOException {
		String v;
		try {
			v = get(pname);
		} catch (Exception e) {
			v = null;
		}
		if (v == null)
			throw new IOException("No such property: " + pname);
		try {
			return Double.valueOf(v).doubleValue();
		} catch (NumberFormatException e) {
			throw new IOException(
				"Double number format exception for " + pname + ": " + v);
		}
	}

	public void setOfftmpl(String off) {
		offtmpl_ = makePath(off);
	} // method setOfftmpl

	public String getOfftmpl() {
		return offtmpl_;
	} // method getOfftmpl

    public String getCacheFile() {
	return cacheFile_;
    }

	/**
	 * Check whether a file path looks secure
	 */
	public static boolean isSecurePath(String path) {
		if (path == null)
			return false;
		path = path.replace(File.separatorChar, '/');

		// No wildcard nor system-interpretable character
		if (path.indexOf('*') >= 0
			|| path.indexOf('?') >= 0
			|| path.indexOf('~') >= 0
			|| path.indexOf('[') >= 0
			|| path.indexOf(']') >= 0
			|| path.indexOf('\'') >= 0
			|| path.indexOf('`') >= 0
			|| path.indexOf('$') >= 0
			|| path.indexOf("..") >= 0)
			return false;
		return true;
	}

	public static boolean isSecurePathStrong(String path, String updir) {
		if (isSecurePath(path)) {
			if (path.startsWith(updir))
				return true;
			return !path.startsWith("/");
		} else
			return false;
	}

	public String getUploadDir() {
		return (uploaddir_ == null ? null : makePath(uploaddir_.trim()));
	}

	public int getJdbcTxIsolation() {
		return txIsolation_;
	}

	private int decodeTxIsolation(String level) {
		if (level == null)
			return AppConfiguration.INT_DEFAULT;
		level = level.trim().toUpperCase();
		if (level.equals("NONE")) {
			return Connection.TRANSACTION_NONE;
		} else if (level.equals("READ_COMMITTED")) {
			return Connection.TRANSACTION_READ_COMMITTED;
		} else if (level.equals("READ_UNCOMMITTED")) {
			return Connection.TRANSACTION_READ_UNCOMMITTED;
		} else if (level.equals("REPEATABLE_READ")) {
			return Connection.TRANSACTION_REPEATABLE_READ;
		} else if (level.equals("SERIALIZABLE")) {
			return Connection.TRANSACTION_SERIALIZABLE;
		}
		return AppConfiguration.INT_DEFAULT;
	}
}
